#include <linux/errno.h>    /* Need this for ENODEOV, etc. */
#include <linux/kernel.h>   /* Need this for printk()  */
#include <linux/major.h>    /* Need this for CQ_MAJOR  */
#include <linux/sched.h>    /* Need this for waking and sleeping */
#include <asm/uaccess.h>    /* Need this for copy_from_user() */ 

#include <linux/cq.h>       /* Note the cq file.  */

void cq_timer_handler(unsigned long);
static void cq_set_pitch(unsigned long);
static void cq_set_tempo(unsigned long);

void speaker_on(void);
void speaker_off(void);

char * morse_code[40] ={ "",\
 "._","_...","_._.","_..",".",".._.","__.","....","..",".___","_._","._..",\
 "__","_.","___",".__.","__._","._.","...","_",".._","..._",".__","_.._",\
 "_.__","__..","_____",".____","..___","...__","...._",".....","_....",\
 "__...","___..","____.","__..__","_._._.","..__.."};

/* The empty string, followed by the 26 letter codes, 
   followed by the 10 numeral codes, followed by the comma, 
   period, and question mark. */

struct cq_struct cq_area =  { 0, 0, 0, {0}, 0, 0, "", NULL};


struct timer_list cq_timer = {NULL, NULL, 0, 0, &cq_timer_handler};

inline char * mcodestring(int asciicode)
{
	char * mc;

	if(asciicode > 122)        /*  Past 'z'  */
		mc = morse_code[CQ_DEF_CHAR];
	else if(asciicode > 96)    /*  Upper case */
		mc = morse_code[asciicode - 96];
	else if(asciicode > 90)    /*  Uncoded punctiation */
		mc = morse_code[CQ_DEF_CHAR];
	else if (asciicode > 64)   /*  Lower case */
		mc = morse_code[asciicode - 64];
	else if (asciicode == 63)   /*  Question Mark */
		mc = morse_code[39];  /* 36 + 3 */
	else if(asciicode > 57)    /*  Uncoded punctiation */
		mc = morse_code[CQ_DEF_CHAR];
	else if (asciicode > 47)   /*  Numerals  */
		mc = morse_code[asciicode - 21]; /* 27 + (asciicode - 48) */
	else if (asciicode == 46)   /*  Period */
		mc = morse_code[38];  /* 36 + 2 */
	else if (asciicode == 44)   /*  Comma */
		mc = morse_code[37]; /* 36 + 1 */
	else
		mc = morse_code[CQ_DEF_CHAR];
	return mc;
}


static int cq_reset(void)
{
	speaker_off();
	cq_set_pitch(CQ_DEF_PITCH);  
	cq_set_tempo(CQ_DEF_TEMPO);  /* Quarter-note */

	return 0;   
}

void cq_timer_handler(unsigned long x)
{
	int status;
	unsigned int time_value;

	time_value = jiffies + 1;

	status = cq_area.flags ;

	/* Assume no more time needed */

	if (status & CQ_ABORT) 
	{
		speaker_off();
		return;
	}
	if ((*(cq_area.mcode) == 0) &&(status & CQ_OFF) 
		&& (cq_area.buffer_offset  + 1 == cq_area.buffer_end))  
				/* This buffer load is done */ 
	{
			wake_up_interruptible(&(cq_area.cq_wait_q));
			return;
	}
	if ((*(cq_area.mcode) == 0) &&(status & CQ_OFF))  
			/* This character is done */ 
	{
		cq_area.buffer_offset = cq_area.buffer_offset + 1;
		x = cq_area.buffer[cq_area.buffer_offset];
		cq_area.mcode = mcodestring(x);

	}
	if ((*(cq_area.mcode) == 0) &&(status & CQ_OFF))  
			/* Previous code put in the empty string */
			/* Empty string gets a time value.  */
	{
		time_value = cq_area.wholenote >> 2;
	} 
	else if (status & CQ_OFF)  /* More dots or dashes left */
	{
		if(*(cq_area.mcode) == '_')
			time_value = cq_area.wholenote >> 2;
		else if (*(cq_area.mcode) == '.')
			time_value = cq_area.wholenote >> 4;
		(cq_area.mcode) = 	(cq_area.mcode)  + 1;

		speaker_on();
	} 
	else if(*(cq_area. mcode) == 0) /* None left but speaker still on */
	{
		time_value = (cq_area.wholenote >> 3) + (cq_area.wholenote >> 2);
						/* inter-letter rest */
		speaker_off();
	} 
	else                    /* More dots or dashes left */ 
	{
		time_value = (cq_area.wholenote >> 3);
				/* Short intra-letter rest */
		speaker_off();
	}
	
	cq_timer.expires = jiffies + time_value;
/*
	cq_timer.expires = time_value;  
		In case you have seen this elsewhere, please note:
		This does Not Work!
*/
	add_timer(&cq_timer);

	return;

}


static ssize_t cq_write(struct file * file, const char * buf, 
				size_t count, loff_t *ppos )
{
	unsigned long copy_size;
	unsigned int time_value;
	ssize_t total_bytes_written = 0;
	time_value = 1;

	total_bytes_written = 0;
	do
	{
		if(signal_pending(current))
		{
			cq_area.flags = (cq_area.flags | CQ_ABORT);
			return total_bytes_written;
		}
		copy_size = (count <= CQ_BUFFER_SIZE ? count : CQ_BUFFER_SIZE);
		cq_area.buffer_offset = 0;
		cq_area.buffer_end = copy_size;
		copy_from_user(cq_area.buffer, buf, copy_size);
		cq_area.mcode = mcodestring(cq_area.buffer[0]);

		time_value = cq_area.wholenote>> 4;
		cq_timer.expires = jiffies + time_value;
		add_timer(&cq_timer);
		speaker_off();
		interruptible_sleep_on(&(cq_area.cq_wait_q));
		total_bytes_written = total_bytes_written + copy_size;
		count = count - copy_size;
		buf = buf + copy_size;
	}while (count > 0); 	
	del_timer(&cq_timer);
	return total_bytes_written;
}


static long long cq_lseek(struct file * file, long long offset, int origin)
{
	return -ESPIPE;
}

static int cq_open(struct inode * inode, struct file * file)
{
	unsigned int minor = MINOR(inode->i_rdev);

	if (minor > 0 )         /*  Stereo No */
		return -ENXIO;
	if ((cq_area.flags & CQ_EXIST) == 0)
		return -ENXIO;
	if (cq_area.flags & CQ_BUSY)
		return -EBUSY;

	cq_area.flags = (cq_area.flags | CQ_BUSY);
	return 0;
}

static int cq_release(struct inode * inode, struct file * file)
{
	cq_reset();

	cq_area.flags = (cq_area.flags & ~CQ_BUSY);
	return 0;
}

static void cq_set_pitch(unsigned long cps)
{
	cq_area.timerdiv =  1193180 / cps;

	/* Frequency of the OSC signal on the AT bus */

}

static void cq_set_tempo(unsigned long beats_per_minute)
{
	/* Quarter note = 1 beat */

	cq_area.wholenote = 24000 / beats_per_minute;

	/* Timer works in centiseconds */

}

static int cq_ioctl(struct inode *inode, struct file *file,
		    unsigned int cmd, unsigned long arg)
{
	unsigned int minor = MINOR(inode->i_rdev);
	int retval = 0;
/*
#ifdef CQ_DEBUG
	printk(KERN_DEBUG "cq%d ioctl, cmd: 0x%x, arg: 0x%x\n", minor, cmd, arg);
#endif
*/
	if (minor > 0)
		return -ENODEV;
	if (((cq_area.flags) & CQ_EXIST) == 0)
		return -ENODEV;
	switch ( cmd ) 
	{
		case CQABORT:
			if (arg)
				cq_area.flags |= CQ_ABORT;
			else
				cq_area.flags &= ~CQ_ABORT;
			break;
		case CQSETPITCH:
			if (arg)
				cq_set_pitch(arg);
			else
				retval = -EINVAL;
			break;
		case CQSETTEMPO:
			if (arg)
				cq_set_tempo(arg);
			else
				retval = -EINVAL;
			break;
		case CQRESET:
			cq_reset();
			break;
 		case CQGETFLAGS:
 			retval = verify_area(VERIFY_WRITE, (void *) arg,
 			    sizeof(int));
 		    if (retval)
 		    		return retval;
 			else 
 			{
 				int status = cq_area.flags;
				copy_to_user((int *) arg, &status, sizeof(int));
			}
			break;
		default:
			retval = -EINVAL;
	}
	return retval;
}


static struct file_operations cq_fops = {
	cq_lseek,
	NULL,		/* cq_read */
	cq_write,	/* cq_write */
	NULL,		/* cq_readdir */
	NULL,		/* cq_poll */
	cq_ioctl,	/* cq_ioctl */
	NULL,		/* cq_mmap */
	cq_open,
	NULL,		/* cq_flush */
	cq_release
};


int cq_init(void)
{
	if (register_chrdev(CQ_MAJOR,"cq",&cq_fops)) 
	{
		printk(KERN_INFO "cq: unable to get major %d\n", CQ_MAJOR);
		return -EIO;
	}

	cq_area.flags = (CQ_OFF | CQ_EXIST);
	cq_reset();

	return 0;
}

